/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/*
 * @headerfile  feature_tracker.h
 *
 * @brief       Header file for the modalcv feature tracker
 *
 * @author      Thomas Patton (thomas.patton@modalai.com) 
 * @date        2024 
 */

#ifndef MODALCV_FEATURE_TRACKER_H
#define MODALCV_FEATURE_TRACKER_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <modal_pipe.h>

#include "pyr_fpx.h"
#include "opt_flow.h"

#define MAX_N_FEATURES         9999

/**
 * @brief Generic handle for feature tracking in modalcv
 * 
 * @ingroup mcv_cvp_types
 */
typedef void* mcv_cvp_feature_tracker_handle;

/**
 * @brief Configuration for feature tracking
 *
 * In its current iteration, the feature tracker uses a pyr_fpx module to get 
 * feature points and and optic flow module to track them. Therefore, we 
 * configure those modules with their respective configuration objects
 */
typedef struct mcv_cvp_feature_tracker_config_t {
    mcv_cvp_pyr_fpx_config_t pyr_fpx_config;
    mcv_cvp_opt_flow_config_t opt_flow_config;
} mcv_cvp_feature_tracker_config_t;

/**
 * @brief Structure fo the output of the feature tracker. 
 */
typedef struct mcv_cvp_feature_tracker_output_t {
    int64_t id; /**< unique id for the feature point, should correspond to a match in the previous frame */
    float x; /**< sub-pixel refined x coord of feature */
    float y; /**< sub-pixel refined y coord of feature */
    float x_prev; /**< previous x location of feature */
    float y_prev; /**< previous y location of feature */
    unsigned char descriptor[32] = {0}; /**< feature descriptor */
    int32_t age; /**< number of frames the feature has been "alive" */
    int8_t score; /**< [0-100] score of the feature */
    int8_t pyr_lvl_mask; /**< which pyramid levels the feature is present on */
    int32_t cvp_scores[3]; /**< the CVP scores, per level */ 
    int8_t reserved_1;
    uint32_t reserved_2;
} __attribute__((packed)) mcv_cvp_feature_tracker_output_t;

/**
 * @brief Initialize the feature tracker with a given config
 * 
 * @param config A populated configuration object
 * @return mcv_cvp_feature_tracker_handle 
 */
mcv_cvp_feature_tracker_handle mcv_cvp_feature_tracker_init(
    mcv_cvp_feature_tracker_config_t config
);

/**
 * @brief De-initialize the feature tracker 
 * 
 * @param handle The handle obtained from the initialization function
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_feature_tracker_deinit(
    mcv_cvp_feature_tracker_handle handle
);

/**
 * @brief Finds and tracks features from image to image
 * 
 * @param handle The image generated by the init function
 * @param img The current image
 * @param output_size The size of the output
 * @param output The output structure
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_feature_tracker_process(
    mcv_cvp_feature_tracker_handle handle, 
    uint8_t* img,
    int16_t* output_size,
    mcv_cvp_feature_tracker_output_t output[MAX_N_FEATURES]
);

#ifdef __cplusplus
}
#endif

#endif // MODALCV_FEATURE_TRACKER_H
